@@ -4,9 +4,10 @@ import sys
 import os.path
 from datetime import datetime
 from dateutil import parser as dateparser
-from dateutil.parser._parser import ParserError as DateParserError
-from traceback import format_exc
+from dateutil.parser import ParserError as DateParserError
+from traceback import format_exc, print_exc
 from io import BytesIO
+import warnings
 
 from PySide2.QtCore import *
 from PySide2.QtWidgets import *
@@ -49,16 +50,19 @@ def array_from_file(filename, date_format=None, dateparser_kw=dateparser_kw):
     Parameters
     ----------
     filename : str
+        The path to the file to be loaded.
     date_format : str or None
         If None, `dateutil.parser.parse` is used, which automatically
-        guesses the format (but uses `dateparserkw` to figure out
+        guesses the format (but uses `dateparser_kw` to figure out
         ambiguous dates).
         Otherwise, the str is used in `datetime.datetime.strptime`
         for faster loading.
     dateparser_kw : dict
         Keyword options for `dateutil.parser.parse`.
 
-    If the filename ends with ".csv" and the first line contains a ","
+    Notes
+    -----
+    If the filename ends with `.csv` and the first two lines contain a `,`
     lines will be split on commas only, otherwise on any combination
     of commas or whitespace.
 
@@ -67,13 +71,14 @@ def array_from_file(filename, date_format=None, dateparser_kw=dateparser_kw):
     the whole line is assumed to be a header and discarded.
 
     If `date_format` is specified, `datetime.datetime.strptime` is used
-    instead to parse dates.
+    instead to parse dates, which can run significantly faster
+    (31 vs 142 seconds on a 86 MB file).
     """
     data = []
     with open(filename) as file:
         if date_format:
             if filename.endswith(".csv"):
-                line = file.readline()
+                line = file.readline() + file.readline()
                 file.seek(0)
                 if "," in line:
                     for line in file:
@@ -84,7 +89,7 @@ def array_from_file(filename, date_format=None, dateparser_kw=dateparser_kw):
                             except ValueError:
                                 try:
                                     row.append(mpldates.date2num(datetime.strptime(val, date_format)))
-                                except DateParserError:
+                                except ValueError:
                                     break
                         if len(row) > 1:
                             data.append(row)
@@ -98,7 +103,7 @@ def array_from_file(filename, date_format=None, dateparser_kw=dateparser_kw):
                     except ValueError:
                         try:
                             row.append(mpldates.date2num(datetime.strptime(val, date_format)))
-                        except DateParserError:
+                        except ValueError:
                             break
                 if len(row) > 1:
                     data.append(row)
@@ -138,15 +143,17 @@ def array_from_file(filename, date_format=None, dateparser_kw=dateparser_kw):
 
 
 class ParserOptionsDialog(QDialog):
-    date_formats_kw = [{"dayfirst":False, "yearfirst": True},   # YMD
-                       {"dayfirst":True, "yearfirst": False},   # DMY
-                       {"dayfirst":False, "yearfirst": False},  # MDY
-                       {"dayfirst":True, "yearfirst": True}]    # YDM
+    date_formats_kw = [{"dayfirst": False, "yearfirst": True},   # YMD
+                       {"dayfirst": True, "yearfirst": False},   # DMY
+                       {"dayfirst": False, "yearfirst": False},  # MDY
+                       {"dayfirst": True, "yearfirst": True}]    # YDM
     def __init__(self, *args, **kwargs):
-        super().__init__(*args, **kwargs, windowFlags=Qt.WindowCloseButtonHint)  # hide ? button
+        super().__init__(*args, **kwargs)
         self.ui = Ui_ParserOptionsDialog()
         self.ui.setupUi(self)
 
+        self.setWindowFlags(self.windowFlags() & ~Qt.WindowContextHelpButtonHint)  # hide ? button
+
         try:
             self.ui.date_format.setCurrentIndex(self.date_formats_kw.index(dateparser_kw))
         except ValueError:
@@ -159,14 +166,22 @@ class ParserOptionsDialog(QDialog):
 
 
 fig = Figure()
+"""
+Instance of `matplotlib.figure.Figure`. Can be used to pre-configure the plot,
+just like when using ``fig, ax = plt.subplots()``
+"""
 ax = fig.add_subplot()
-
+"""
+A subplot of `fig`. Can be used to pre-configure the plot,
+just like when using ``fig, ax = plt.subplots()``, or removed and replaced
+by a grid of axes using `fig.add_subplot()`
+"""
 
 _files_to_load = []
 _data_to_load = []
 
 suppress_errors = False
-"""bool: Whether to hide errors and other output from add_file and add_array"""
+"""bool: Whether to hide errors and other output from `add_file` and `add_array.`"""
 
 
 def add_file(filename):
@@ -176,7 +191,7 @@ def add_file(filename):
     Parameters
     ----------
     filename : str
-        The file to be loaded (should be 2d).
+        The file to be loaded.
 
     Returns
     -------
@@ -185,14 +200,13 @@ def add_file(filename):
 
     Notes
     -----
-        `filename` should have at least 2 rows and 2 columns.
+        `filename` should have at least 2 rows and 2 columns of data.
     """
-    if os.path.isfile(filename):
-        _files_to_load.append(filename)
-        return True
-    if not suppress_errors:
+    if not suppress_errors and not os.path.isfile(filename):
         print(f"[PloPy]: '{filename}' is not a file.")
-    return False
+        return False
+    _files_to_load.append(filename)
+    return True
 
 
 def add_array(array, name=None, raise_=False):
@@ -201,10 +215,13 @@ def add_array(array, name=None, raise_=False):
 
     Parameters
     ----------
-    array : np.array or list-like of list-like of float
-        The data to be loaded (should be 2d).
+    array : np.array or iterable of iterable of float
+        The data to be loaded.
     name : str or None
         The name to be used for the tab.
+    raise : bool
+        Whether to raise the exception if one is caused
+        by converting the input to a numpy array.
 
     Returns
     -------
@@ -284,8 +301,8 @@ class FigureCanvasCustom(FigureCanvas):
 
 
 class ViewArrayDialog(QDialog):
-    def __init__(self, data, name, *args, **kwargs):
-        super().__init__(*args, **kwargs, windowTitle=f"Plo.Py - View {name}")
+    def __init__(self, parent, data, name, *args, **kwargs):
+        super().__init__(parent, *args, **kwargs, windowTitle=f"Plo.Py - View {name}")
         self.ui = Ui_ViewArrayDialog()
         self.ui.setupUi(self)
         self.data = data
@@ -308,9 +325,9 @@ class ViewArrayDialog(QDialog):
 
 
 class ViewFileDialog(ViewArrayDialog):
-    def __init__(self, data, filename, *args, **kwargs):
+    def __init__(self, parent, data, filename, *args, **kwargs):
         self.__has_loaded_raw = False
-        QDialog.__init__(self, *args, **kwargs, windowTitle=f"Plo.Py - View {os.path.basename(filename)}")
+        QDialog.__init__(self, parent, *args, **kwargs, windowTitle=f"Plo.Py - View {os.path.basename(filename)}")
         self.ui = Ui_ViewFileDialog()
         self.ui.setupUi(self)
         self.data = data
@@ -430,7 +447,7 @@ class LineOptions(QWidget):
     def update_color(self, dialog=True):
         color = to_hex(self.line.get_c())
         if dialog:
-            new_qcolor = QColorDialog.getColor(color)
+            new_qcolor = QColorDialog.getColor(color, self)
             if not new_qcolor.isValid():
                 return
             color = new_qcolor.name()  # to hex
@@ -439,7 +456,7 @@ class LineOptions(QWidget):
         self.ui.color_frame.setStyleSheet(f"#color_frame {{background-color: {color}}}")
 
     def update_axes(self, text):
-        new_axes = next((axes for axes in fig.axes if axes_to_str(axes) == text), None)
+        new_axes = next((axes for axes in fig.axes if axes.plopy_name == text), None)
         if not new_axes:
             print(self.line, "could not be moved to a new axes.")
             return
@@ -459,13 +476,15 @@ class ArrayPlotDialog(QDockWidget):
         cls._title_count += 1
         return f"Array {cls._title_count}"
 
-    def __init__(self, data, axes_list, windowTitle=None, *args, add_initial_line=True, **kwargs):
-        assert len(data.shape) == 2  #  input data must be a 2D array
+    def __init__(self, parent, data, axes_list, windowTitle=None, *args, add_initial_line=True, **kwargs):
+        if len(data.shape) != 2:
+            raise ValueError(f"data.shape should have length 2, instead was {data.shape}")
+        self.parent = parent
         self.data = data
         self.axes_list = axes_list  # synced to update the line's axes chooser
         if not windowTitle:
             windowTitle = self.generate_title()
-        super().__init__(*args, **kwargs, windowTitle=windowTitle)
+        super().__init__(parent, *args, **kwargs, windowTitle=windowTitle)
         self.ui = Ui_PlotOptions()
         self.ui.setupUi(self)
 
@@ -503,7 +522,7 @@ class ArrayPlotDialog(QDockWidget):
         self.current_line_number += 1
 
     def detail_dialog(self):
-        ViewArrayDialog(self.data, self.windowTitle()).exec_()
+        ViewArrayDialog(self.parent, self.data, self.windowTitle()).exec_()
 
     def show(self):
         super().show()
@@ -517,12 +536,12 @@ class ArrayPlotDialog(QDockWidget):
 
 
 class FilePlotDialog(ArrayPlotDialog):
-    def __init__(self, filename, axes_list, date_format=None, *args, **kwargs):
-        super().__init__(array_from_file(filename, date_format), axes_list, *args, **kwargs, windowTitle=os.path.basename(filename))
+    def __init__(self, parent, filename, axes_list, date_format=None, *args, **kwargs):
+        super().__init__(parent, array_from_file(filename, date_format), axes_list, *args, **kwargs, windowTitle=os.path.basename(filename))
         self.filename = filename
 
     def detail_dialog(self):
-        ViewFileDialog(self.data, self.filename).exec_()
+        ViewFileDialog(self.parent, self.data, self.filename).exec_()
 
 
 class AddAxesDialog(QDialog):
@@ -534,12 +553,21 @@ class AddAxesDialog(QDialog):
 
         self.ui.nrows.valueChanged.connect(self.update_max_index)
         self.ui.ncols.valueChanged.connect(self.update_max_index)
-        self.update_max_index()
+        self.ui.index.valueChanged.connect(self.update_radiobutton_state)
 
     def update_max_index(self):
         self.ui.index.setMaximum(self.ui.nrows.value() * self.ui.ncols.value())
+        self.update_radiobutton_state()
+
+    def update_radiobutton_state(self):
+        location = self.ui.nrows.value(), self.ui.ncols.value(), self.ui.index.value()
+        exists = f"Axes {location}" in (axes.plopy_name for axes in fig.axes)
+        self.ui.radio_new.setEnabled(not exists)
+        self.ui.radio_twinx.setEnabled(exists)
+        self.ui.radio_twiny.setEnabled(exists)
 
     def show(self):
+        self.update_max_index()
         super().show()
         self.raise_()
         self.activateWindow()
@@ -591,7 +619,7 @@ class AxesLimDialog(QDialog):
 
     def show(self, ax):
         self.current_ax = ax
-        self.ui.axes_box.setText(axes_to_str(ax))
+        self.ui.axes_box.setText(ax.plopy_name)
         self.update_from_axes()
 
         super().show()
@@ -628,9 +656,15 @@ class AxesOptions(QWidget):
         #self.ui.legend_loc.insertSeparator(9)
 
     def setup(self, _=None):
-        self.ui.title.setText(self.axes.get_title())
-        self.ui.x_label.setText(self.axes.get_xlabel())
-        self.ui.y_label.setText(self.axes.get_ylabel())
+        if "Twin" in self.axes.plopy_name:
+            if "TwinX" in self.axes.plopy_name:
+                self.ui.x_gbox.setEnabled(False)
+                self.ui.y_label.setText(self.axes.get_ylabel())
+            else:
+                self.ui.y_gbox.setEnabled(False)
+                self.ui.x_label.setText(self.axes.get_xlabel())
+        else:
+            self.ui.title.setText(self.axes.get_title())
 
     def set_title(self, text):
         self.axes.set_title(text)
@@ -653,7 +687,7 @@ class AxesOptions(QWidget):
             l = mpldates.AutoDateLocator()
             axis.set_major_locator(l)
             axis.set_major_formatter(mpldates.AutoDateFormatter(l))
-        elif selection == "None":
+        elif selection == "Blank":
             axis.set_major_locator(ticker.NullLocator())
 
         if selection.endswith("minor ticks"):
@@ -662,24 +696,34 @@ class AxesOptions(QWidget):
             axis.set_minor_locator(ticker.NullLocator())
 
     def set_xticks(self, selection):
-        self.set_axis_ticks(self.axes.xaxis, selection)
-
         try:
+            if selection.endswith("Logarithmic"):
+                self.set_axis_ticks(self.axes.xaxis, "Normal")
+                self.axes.set_xscale('log' if selection == "Logarithmic" else 'symlog')
+            else:
+                self.axes.set_xscale('linear')
+                self.set_axis_ticks(self.axes.xaxis, selection)
+
             update()
         except Exception as e:
             print("Error when configuring x ticks:", e)
-            if selection != "None":
-                self.ui.x_tick_type.setCurrentText("None")  # note: this will re-run this function
+            if selection != "Blank":
+                self.ui.x_tick_type.setCurrentText("Blank")  # note: this will re-run this function
 
     def set_yticks(self, selection):
-        self.set_axis_ticks(self.axes.yaxis, selection)
-
         try:
+            if selection.endswith("Logarithmic"):
+                self.set_axis_ticks(self.axes.yaxis, "Normal")
+                self.axes.set_yscale('log' if selection == "Logarithmic" else 'symlog')
+            else:
+                self.axes.set_yscale('linear')
+                self.set_axis_ticks(self.axes.yaxis, selection)
+
             update()
         except Exception as e:
             print("Error when configuring y ticks:", e)
-            if selection != "None":
-                self.ui.y_tick_type.setCurrentText("None")  # note: this will re-run this function
+            if selection != "Blank":
+                self.ui.y_tick_type.setCurrentText("Blank")  # note: this will re-run this function
 
     def set_tick_direction(self, index):
         self.axes.tick_params(which='both', direction=('out', 'in', 'inout')[index])
@@ -717,7 +761,7 @@ class FigureDialog(QDockWidget):
         self.ui.title.textChanged.connect(self.set_title)
         self.ui.tight_layout.stateChanged.connect(self.set_layout)
 
-        self.axes_lim_dialog = AxesLimDialog()
+        self.axes_lim_dialog = AxesLimDialog(mainwindow)
 
         for axes in figure.axes:
             self.add_axes(axes)
@@ -742,7 +786,7 @@ class FigureDialog(QDockWidget):
         widget = self.ui.tab_widget.widget(index)
         if widget is not None:
             n_lines = len(widget.axes.lines)
-            name = axes_to_str(widget.axes)
+            name = widget.axes.plopy_name
             if n_lines and QMessageBox.Ok != QMessageBox.warning(self, "Plo.Py - Remove Axes?",
                                f"Are you sure you want to delete {name}?\nIt still has {n_lines} line{'s' if n_lines != 1 else ''} attached.",
                                QMessageBox.Ok | QMessageBox.Cancel):
@@ -755,6 +799,23 @@ class FigureDialog(QDockWidget):
 
     def add_axes(self, axes):
         name = axes_to_str(axes)
+        if self.axes_list.findItems(name):  # axes already exists at that location, may be a twin
+            for other_ax in axes.get_shared_x_axes().get_siblings(axes):
+                if other_ax is axes:
+                    continue
+                if other_ax.bbox.bounds == axes.bbox.bounds:
+                    name += " TwinX"
+                    break
+            else:
+                for other_ax in axes.get_shared_y_axes().get_siblings(axes):
+                    if other_ax is axes:
+                        continue
+                    if other_ax.bbox.bounds == axes.bbox.bounds:
+                        name += " TwinY"
+                        break
+            while self.axes_list.findItems(name):
+                name += " copy"
+        axes.plopy_name = name
 
         self.axes_list.appendRow(QStandardItem(name))  # append axes name (updates all ComboBoxes)
 
@@ -765,11 +826,21 @@ class FigureDialog(QDockWidget):
     def create_axes(self):
         d_ui = self.add_axes_dialog.ui
         location = d_ui.nrows.value(), d_ui.ncols.value(), d_ui.index.value()
-        if self.axes_list.findItems(f"Axes {location}"):
-            print(f"Axes already exists at {location}, using that instead")
-            qapp.beep()
-            return
-        self.add_axes(self.figure.add_subplot(*location)).setup()
+        for axes in fig.axes:
+            if axes.plopy_name == f"Axes {location}":
+                if d_ui.radio_twinx.isChecked():
+                    new_ax = axes.twinx()
+                    break
+                elif d_ui.radio_twiny.isChecked():
+                    new_ax = axes.twiny()
+                    break
+                else:
+                    print(f"Axes already exists at {location}, using that instead")
+                    qapp.beep()
+                    return
+        else:
+            new_ax = self.figure.add_subplot(*location)
+        self.add_axes(new_ax).setup()
         update()
 
     def show(self):
@@ -791,6 +862,15 @@ class MainWindow(QMainWindow):
         self.ui = Ui_MainWindow()
         self.ui.setupUi(self)
 
+        _showwarning_old = warnings.showwarning
+        def show_warning_new(message, category, filename, lineno, file=None, line=None):
+            try:
+                self.ui.statusbar.showMessage(f"Warning: {message}", 20000)
+            except Exception:
+                print_exc()
+            _showwarning_old(message, category, filename, lineno, file, line)
+        warnings.showwarning = show_warning_new
+
         self.icon = QIcon()
         icon_path = os.path.join(os.path.dirname(__file__), "ui", "icon_")
         self.icon.addFile(icon_path+"16x16.png", QSize(16, 16))
@@ -825,15 +905,18 @@ class MainWindow(QMainWindow):
         self.ui.action_docs.triggered.connect(lambda: QDesktopServices.openUrl(self.ui.action_docs.toolTip()))
 
         self.show()
+        self.raise_()
+        self.activateWindow()
+
         qapp.processEvents()
-        self.adv_load_dialog = AdvancedLoadDialog()
+        self.adv_load_dialog = AdvancedLoadDialog(self)
         self.figoptions = FigureDialog(self, self.axes_list)
         self.ui.action_figure_options.triggered.connect(self.figoptions.show)
         nav.edit_parameters = self.figoptions.show
 
         self.add_existing_lines()
-        self.add_data_files(_files_to_load)
         self.add_arrays(_data_to_load)
+        self.add_data_files(_files_to_load, always_alert=True)
 
     def choose_data_files(self, date_format=None):
         filenames, extension = QFileDialog.getOpenFileNames(self,
@@ -845,7 +928,7 @@ class MainWindow(QMainWindow):
         if result == QDialog.Accepted:
             self.choose_data_files(str(self.adv_load_dialog.ui.date_format.currentText()) or None)
 
-    def add_data_files(self, filenames, date_format=None):
+    def add_data_files(self, filenames, date_format=None, always_alert=False):
         if filenames:
             pbar = QProgressBar(maximum=len(filenames)+1, alignment=Qt.AlignCenter)
             self.ui.statusbar.addPermanentWidget(pbar)
@@ -859,7 +942,7 @@ class MainWindow(QMainWindow):
                 qapp.processEvents()
 
                 try:
-                    dialog = FilePlotDialog(filename, self.axes_list, date_format, parent=self)
+                    dialog = FilePlotDialog(self, filename, self.axes_list, date_format)
                 except Exception:
                     # errors will be shown after all files have been attempted
                     error_messages.append(f"Error when loading file '{filename}'\n\n{format_exc()}")
@@ -878,10 +961,13 @@ class MainWindow(QMainWindow):
             for msg in error_messages:
                 QMessageBox.warning(self, "Plo.Py", msg)
 
+        elif always_alert:
+            qapp.alert(self)
+
     def add_arrays(self, arrays):
         show_functions = []
         for array, name in arrays:
-            dialog = ArrayPlotDialog(array, self.axes_list, windowTitle=name, parent=self)
+            dialog = ArrayPlotDialog(self, array, self.axes_list, windowTitle=name)
             self.add_plot_dialog(dialog)
             show_functions.append(dialog.show)
 
@@ -891,7 +977,7 @@ class MainWindow(QMainWindow):
     def add_existing_lines(self):
         for axes in fig.axes:
             for line in axes.lines:
-                dialog = ArrayPlotDialog(line.get_xydata(), self.axes_list, add_initial_line=False, windowTitle=line.get_label(), parent=self)
+                dialog = ArrayPlotDialog(self, line.get_xydata(), self.axes_list, add_initial_line=False, windowTitle=line.get_label())
                 dialog.add_line(line)
                 self.add_plot_dialog(dialog)
 
@@ -910,7 +996,7 @@ class MainWindow(QMainWindow):
         qapp.clipboard().setImage(QImage.fromData(buffer.getvalue()))
 
     def configure_parser(self):
-        d = ParserOptionsDialog()
+        d = ParserOptionsDialog(self)
         d.exec_()
         d.deleteLater()
 
